/*

(Yet Another) Tiny Scheme

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tscm.hpp"

class FILEInputStream : public TSInputStream
{
public:

    FILEInputStream(FILE * f, char * fn);
    virtual TSCh GetCh(int * eof);
    virtual void UnGetCh(TSCh ch) {ungetc(ch, File);}
    virtual TSCh * GetFilename() {return(Filename);}
    virtual int GetLineNumber() {return(LineNumber);}

private:

    FILE * File;
    int LineNumber;
    TSCh * Filename;
};

FILEInputStream::FILEInputStream(FILE * f, char * fn)
{
    int fnl = strlen(fn) + 1;
    int fdx;

    File = f;
    LineNumber = 1;
    Filename = new TSCh[fnl];
    for (fdx = 0; fdx < fnl; fdx++)
        Filename[fdx] = fn[fdx];
}

TSCh FILEInputStream::GetCh(int * eof)
{
    int ch;

    ch = fgetc(File);
    *eof = (ch == EOF ? 1 : 0);
    if (ch == '\n')
        LineNumber += 1;
    return(ch);
}

class FILEOutputStream : public TSOutputStream
{
public:

    FILEOutputStream(FILE * f) {File = f;}
    virtual void PutCh(TSCh ch) {fputc(ch, File);}
    virtual void PutString(TSCh * s, int sl);
    virtual void PutString(char * s) {fputs(s, File);}

private:

    FILE * File;
};

void FILEOutputStream::PutString(TSCh * s, int sl)
{
    int sdx;

    for (sdx = 0; sdx < sl; sdx++)
        PutCh(s[sdx]);
}

int Load(TSWorld * wld, char * fn)
{
    FILE * f = fopen(fn, "r");

    if (f == 0)
        return(1);

    FILEInputStream fis(f, fn);
    TSObject * obj;

    for (;;)
    {
        TSLineNumberMap lnm(fis.GetFilename());

        obj = wld->Read(&fis, &lnm);
        if (obj == wld->EndOfFile)
            break;
        obj = wld->Evaluate(obj, &lnm);
        if (obj != wld->Unspecified)
        {
            wld->Write(wld->DefaultOutput, obj);
            wld->DefaultOutput->PutString("\n");
        }
    }

    fclose(f);
    return(0);
}

DefineBuiltin("load", Load, "s")(TSWorld * wld, int argc, TSObject * argv[])
{
    char buf[256];
    int idx;
    TSString * st = AsString(argv[0]);

    for (idx = 0; idx < st->Length; idx++)
        buf[idx] = (char) st->String[idx];
    buf[idx] = 0;

    if (Load(wld, buf))
        wld->ThrowError("Unable to open file", argv[0]);
    return(wld->Unspecified);
}

struct SchemeError
{
    // Nothing.
};

static SchemeError se;

class Application : public TSApplication
{
public:

    virtual void FatalError();
    virtual int Write(TSOutputStream * os, TSObject * obj) {return(0);}
    virtual int Display(TSOutputStream * os, TSObject * obj) {return(0);}
};

void Application::FatalError()
{
    throw se;
}

int main(int argc, char * argv[])
{
    TSObject * obj;
    FILEInputStream fis(stdin, "<stdin>");
    FILEOutputStream fos(stdout);
    TSWorld * wld = new TSWorld(&fos, new Application);
    int adx;

    AddAllBuiltins(wld);

    if (argc > 1)
    {
        try
        {
            for (adx = 1; adx < argc; adx++)
                Load(wld, argv[adx]);
        }
        catch (SchemeError)
        {
            return(1);
        }

        delete wld;
        return(0);
    }

    for (;;)
    {
        try
        {
            TSLineNumberMap lnm(fis.GetFilename());

            fos.PutString("=] ");
            obj = wld->Read(&fis, &lnm);
            if (obj == wld->EndOfFile)
                break;
            obj = wld->Evaluate(obj, &lnm);
            if (obj != wld->Unspecified)
            {
                wld->Write(&fos, obj);
                fos.PutString("\n");
            }
        }
        catch (SchemeError)
        {
            // Nothing.
        }
    }

    delete wld;
    return(0);
}
